/*************************************************
* OpenSSL Modular Exponentiation Source File     *
* (C) 1999-2006 The Botan Project                *
*************************************************/

#include <botan/eng_ossl.h>
#include <botan/bn_wrap.h>

namespace Botan {

namespace {

/*************************************************
* OpenSSL Modular Exponentiator                  *
*************************************************/
class OpenSSL_Modular_Exponentiator : public Modular_Exponentiator
   {
   public:
      void set_base(const BigInt& b) { base = b; }
      void set_exponent(const BigInt& e) { exp = e; }
      BigInt execute() const;
      BigInt reduce_by_modulus(const BigInt&) const;
      Modular_Exponentiator* copy() const
         { return new OpenSSL_Modular_Exponentiator(*this); }

      OpenSSL_Modular_Exponentiator(const BigInt& n) : mod(n) {}
   private:
      OSSL_BN base, exp, mod;
      OSSL_BN_CTX ctx;
   };

/*************************************************
* Compute the result                             *
*************************************************/
BigInt OpenSSL_Modular_Exponentiator::execute() const
   {
   OSSL_BN r;
   BN_mod_exp(r.value, base.value, exp.value, mod.value, ctx.value);
   return r.to_bigint();
   }

/*************************************************
* Reduce this number by the modulus              *
*************************************************/
BigInt OpenSSL_Modular_Exponentiator::reduce_by_modulus(const BigInt& n) const
   {
   OSSL_BN r(n);
   BN_mod(r.value, r.value, mod.value, ctx.value);
   if(n.is_negative())
      {
      OSSL_BN r2;
      BN_sub(r2.value, mod.value, r.value);
      return r2.to_bigint();
      }
   return r.to_bigint();
   }

}

/*************************************************
* Return the OpenSSL-based modular exponentiator *
*************************************************/
Modular_Exponentiator* OpenSSL_Engine::mod_exp(const BigInt& n,
                                           Power_Mod::Usage_Hints) const
   {
   return new OpenSSL_Modular_Exponentiator(n);
   }

}
